@@ -502,7 +502,7 @@ class SASsession():
         enc = self._io.sascfg.encoding
         if enc == '':
            self._io.sascfg.encoding = 'utf_8'
-        res = self.submit(sysvars, "text")['LOG']
+        res = self._io.submit(sysvars, "text")['LOG']
         self._io.sascfg.encoding = enc
 
         vlist         = res.rpartition('SYSSCP=')
@@ -554,7 +554,7 @@ class SASsession():
         self.workpath = self.workpath + self.hostsep
 
         if self.sascfg.autoexec:
-            self.submit(self.sascfg.autoexec)
+            self._io.submit(self.sascfg.autoexec)
 
         if self.sascfg.m5dsbug is None:
            if self.sasver[:9] in ['9.04.01M5'] and self.sascei in ['utf-8', 'euc-cn', 'euc-jp', 'euc-kr', 'shift-jis', 'big5']:
@@ -572,7 +572,7 @@ class SASsession():
         # this is to support parsing the log to fring log records w/ 'ERROR' when diagnostic logging is enabled.
         # in thi scase the log can have prefix and/or suffix info so the 'regular' log data is in the middle, not left justified
         if self.sascfg.mode in ['STDIO', 'SSH', '']:
-           ll = self.submit("""data _null_; file STDERR; put %upcase('col0REG='); 
+           ll = self._io.submit("""data _null_; file STDERR; put %upcase('col0REG='); 
                                data _null_; put %upcase('col0LOG=');run;""", results='text')
            regoff = len(ll['LOG'].rpartition('COL0REG=')[0].rpartition('\n')[2])
            logoff = len(ll['LOG'].rpartition('COL0LOG=')[0].rpartition('\n')[2])
@@ -649,13 +649,13 @@ class SASsession():
 
         return self._io._asubmit(code, results)
 
-    def submitLOG(self, code, results: str = '', prompt: dict = None):
+    def submitLOG(self, code, results: str = '', prompt: dict = None, printto=False):
         '''
         This method is a convenience wrapper around the submit() method. It executes the submit then prints the LOG that was returned.
         '''
-        print(self.submit(code, results, prompt)['LOG'])
+        print(self.submit(code, results, prompt, printto)['LOG'])
 
-    def submitLST(self, code, results: str = '', prompt: dict = None, method: str = None):
+    def submitLST(self, code, results: str = '', prompt: dict = None, method: str = None, printto=False):
         '''
         This method is a convenience wrapper around the submit() method. It executes the submit then renders the LST that was returned,
         as either HTML or TEXT, depending upon results. The method= parameter allows you to adjust what gets returned to suit your needs.
@@ -678,7 +678,7 @@ class SASsession():
            else:
               results = self.results
 
-        ll  = self.submit(code, results, prompt)
+        ll  = self.submit(code, results, prompt, printto)
 
         if results.upper() == 'HTML':
            if   method.lower() == 'listonly':
@@ -705,15 +705,16 @@ class SASsession():
            else:
               print(ll['LOG']+"\n"+ll['LST'])
 
-    def submit(self, code: str, results: str = '', prompt: dict = None) -> dict:
+    def submit(self, code: str, results: str = '', prompt: dict = None, printto=False) -> dict:
         '''
         This method is used to submit any SAS code. It returns the Log and Listing as a python dictionary.
 
-        - code    - the SAS statements you want to execute
-        - results - format of results. 'HTML' by default, alternatively 'TEXT'
-        - prompt  - dict of names:flags to prompt for; create macro variables (used in submitted code), then keep or delete
-                    the keys which are the names of the macro variables. The boolean flag is to either hide what you type and delete the macros,
-                    or show what you type and keep the macros (they will still be available later).
+        :param saspath: overrides saspath Dict entry of cfgname in sascfg_personal.py file
+        :param code:    the SAS statements you want to execute
+        :param results: format of results. 'HTML' by default, alternatively 'TEXT'
+        :param prompt:  dict of names and flags to prompt for; create macro variables (used in submitted code), then keep or delete \
+                        the keys which are the names of the macro variables. The boolean flag is to either hide what you type and \
+                        delete the macros, or show what you type and keep the macros (they will still be available later).
 
             for example (what you type for pw will not be displayed, user and dsname will):
 
@@ -727,7 +728,11 @@ class SASsession():
                              prompt = {'user': False, 'pw': True, 'dsname': False}
                              )
 
-            Returns - a Dict containing two keys:values, [LOG, LST]. LOG is text and LST is 'results' (HTML or TEXT)
+        :param printto: this option, when set to True, will cause saspy to issue a 'proc printto;run;' after the code that is being \
+                        submitted. This will 'undo' any proc printto w/in the submitted code that redirected the LOG or LST, to return \
+                        the LOG/LST back to saspy. This is explained in more detail in the doc: https://sassoftware.github.io/saspy/limitations.html
+
+        :return: a Dict containing two keys:values, [LOG, LST]. LOG is text and LST is 'results' (HTML or TEXT)
 
         NOTE: to view HTML results in the ipykernel, issue: from IPython.display import HTML  and use HTML() instead of print()
 
@@ -749,9 +754,8 @@ class SASsession():
             else:
                 results = self.results
 
-        ll = self._io.submit(code, results, prompt)
+        ll = self._io.submit(code, results, prompt, undo=printto)
 
-        self._lastlog = ll['LOG']
         return ll
 
     def saslog(self) -> str:
@@ -777,8 +781,8 @@ class SASsession():
 
     def teach_me_SAS(self, nosub: bool):
         """
-        nosub - bool. True means don't submit the code, print it out so I can see what the SAS code would be.
-            False means run normally - submit the code.
+        :param nosub: bool. True means don't submit the code, print it out so I can see what the SAS code would be. \
+                      False means run normally - submit the code.
         """
         self.nosub = nosub
 
@@ -1213,8 +1217,8 @@ class SASsession():
 
            - MEMORY the original method. Streams the data over and builds the dataframe on the fly in memory
            - CSV    uses an intermediary Proc Export csv file and pandas read_csv() to import it; faster for large data
-           - DISK   uses the original (MEMORY) method, but persists to disk and uses pandas read to import.
-                    this has better support than CSV for embedded delimiters (commas), nulls, CR/LF that CSV
+           - DISK   uses the original (MEMORY) method, but persists to disk and uses pandas read to import. \
+                    this has better support than CSV for embedded delimiters (commas), nulls, CR/LF that CSV \
                     has problems with 
 
         :param kwargs: a dictionary. These vary per access method, and are generally NOT needed.
@@ -1532,7 +1536,7 @@ class SASsession():
         :param value: python variable, that can be resolved to a string, to use for the value to assign to the macro variable
 
         """
-        ll = self.submit("%let " + name + "=%NRBQUOTE(" + str(value) + ");\n")
+        ll = self._io.submit("%let " + name + "=%NRBQUOTE(" + str(value) + ");\n")
 
     def symget(self, name: str, outtype=None):
         """
@@ -1540,7 +1544,7 @@ class SASsession():
         :param outtype: [optional] desired output type of the python variable; valid types are [int, float, str]
 
         """
-        ll = self.submit("%put " + name + "=&" + name + " "+ name+"END=;\n")
+        ll = self._io.submit("%put " + name + "=&" + name + " "+ name+"END=;\n")
         l2 = ll['LOG'].rpartition(name + "=")[2].rpartition(name+"END=")[0].strip().replace('\n','') 
 
         if outtype is not None and type(outtype) not in [int, float, str]:
@@ -1641,7 +1645,7 @@ class SASsession():
             print(code)
             return None
         else:
-           ll = self.submit(code, results='text')
+           ll = self._io.submit(code, results='text')
 
         librefs = []
         log = ll['LOG'].rpartition('LIBREFSEND=')[0].rpartition('LIBREFSSTART=')
@@ -1704,7 +1708,7 @@ class SASsession():
             print(code)
             return None
         else:
-           ll = self.submit(code, results='text')
+           ll = self._io.submit(code, results='text')
 
         dirlist = []
 
@@ -1725,7 +1729,7 @@ class SASsession():
         return dirlist
 
 
-    def list_tables(self, libref, results: str = 'list') -> list:
+    def list_tables(self, libref: str='work', results: str = 'list') -> list:
         """
         This method returns a list of tuples containing MEMNAME, MEMTYPE of members in the library of memtype data or view
 
@@ -1733,7 +1737,7 @@ class SASsession():
         """
 
         if not self.nosub:
-           ll = self.submit("%put LIBREF_EXISTS=%sysfunc(libref("+libref+")) LIB_EXT_END=;")
+           ll = self._io.submit("%put LIBREF_EXISTS=%sysfunc(libref("+libref+")) LIB_EXT_END=;")
 
            exists = int(ll['LOG'].rpartition('LIBREF_EXISTS=')[2].rpartition('LIB_EXT_END=')[0])
 
@@ -1756,7 +1760,7 @@ class SASsession():
             print(code)
             return None
         else:
-           ll = self.submit(code, results='text')
+           ll = self._io.submit(code, results='text')
 
         if results != 'list':
            res = self.sd2df('_saspy_lib_list', 'work')
@@ -1774,7 +1778,7 @@ class SASsession():
         run;
         """
         
-        ll  = self.submit(code, results='text')
+        ll  = self._io.submit(code, results='text')
         
         log = ll['LOG'].rpartition('MEMEND=')[0].rpartition('MEMSTART=')
                                                                                   
@@ -1800,14 +1804,14 @@ class SASsession():
            code  = "filename "+fileref+" '"+filepath+"';\n"
            code += "%put FILEREF_EXISTS=%sysfunc(fexist("+fileref+")) FILE_EXTEND=;"
    
-           ll = self.submit(code)
+           ll = self._io.submit(code)
    
            exists = int(ll['LOG'].rpartition('FILEREF_EXISTS=')[2].rpartition(' FILE_EXTEND=')[0])
    
            if exists != 1:
               if not quiet:
                  print('The filepath provided does not exist')
-              ll = self.submit("filename "+fileref+" clear;")
+              ll = self._io.submit("filename "+fileref+" clear;")
               return None
 
         if results != 'dict':
@@ -1835,7 +1839,7 @@ class SASsession():
                print(code)
                return None
            else:
-              ll  = self.submit(code, results='text')
+              ll  = self._io.submit(code, results='text')
    
            res = self.sd2df('_SASPY_FILE_INFO', 'work')
            return res
@@ -1871,7 +1875,7 @@ class SASsession():
             print(code)
             return None
         else:
-           ll  = self.submit(code, results='text')
+           ll  = self._io.submit(code, results='text')
 
         vi = len(ll['LOG'].rpartition('INFOEND=')[0].rpartition('\n')[2])                          
  
