@@ -19,11 +19,13 @@ import json
 import os
 import ssl
 import sys
+import urllib
 
 import tempfile as tf
 from time import sleep
 
 from saspy.sasexceptions import (SASHTTPauthenticateError)
+from saspy.sasexceptions import (SASHTTPconnectionError)
 try:
    import pandas as pd
    import numpy  as np
@@ -288,10 +290,12 @@ class SASconfigHTTP:
 
    def _authenticate(self, user, pw):
       #import pdb; pdb.set_trace()
-
+      uuser = urllib.parse.quote(user)
+      upw   = urllib.parse.quote(pw)
+      
       # POST AuthToken
       conn = self.HTTPConn; conn.connect()
-      d1 = ("grant_type=password&username="+user+"&password="+pw).encode(self.encoding)
+      d1 = ("grant_type=password&username="+uuser+"&password="+upw).encode(self.encoding)
       basic = base64.encodebytes("sas.tkmtrb:".encode(self.encoding))
       authheader = '%s' % basic.splitlines()[0].decode(self.encoding)
       headers={"Accept":"application/vnd.sas.compute.session+json","Content-Type":"application/x-www-form-urlencoded",
@@ -418,16 +422,25 @@ class SASsessionHTTP():
       conn = self.sascfg.HTTPConn; conn.connect()
       d1 = '{"name":"'+self.sascfg.ctxname+'", "description":"saspy session", "version":1, "environment":{"options":'+options+'}}'
       headers={"Accept":"application/vnd.sas.compute.session+json","Content-Type":"application/vnd.sas.compute.session.request+json","Authorization":"Bearer "+self.sascfg._token}
-      conn.request('POST', uri, body=d1, headers=headers)
-      req = conn.getresponse()
+
+      try:
+         conn.request('POST', uri, body=d1, headers=headers)
+         req = conn.getresponse()
+      except:
+         #print("Could not acquire a SAS Session for context: "+self.sascfg.ctxname)
+         raise SASHTTPconnectionError(msg="Could not acquire a SAS Session for context: "+self.sascfg.ctxname+". Exception info:\n"+str(sys.exc_info()))
+         #return None
+
       status = req.status
       resp = req.read()
       conn.close()
 
       if status > 299:
-         print("Failure in POST Session \n"+resp.decode(self.sascfg.encoding))
-         print("Could not acquire a SAS Session for context: "+self.sascfg.ctxname)
-         return None
+         #print("Failure in POST Session \n"+resp.decode(self.sascfg.encoding))
+         #print("Could not acquire a SAS Session for context: "+self.sascfg.ctxname)
+         msg="Could not acquire a SAS Session for context: "+self.sascfg.ctxname+". Exception info:\nStatus="+str(status)+"\nResponse="+str(resp)
+         raise SASHTTPconnectionError(msg)
+         #return None
 
       self._session = json.loads(resp.decode(self.sascfg.encoding))
 
@@ -573,7 +586,7 @@ class SASsessionHTTP():
       headers={"Accept":"application/vnd.sas.collection+json", "Authorization":"Bearer "+self.sascfg._token}
       while i < len(results):
          # GET an ODS Result
-         if results[i].get('type') == 'ODS':
+         if results[i].get('type') == 'ODS' and len(results[i].get('links')) > 0:
             conn.request('GET', results[i].get('links')[0].get('href'), headers=headers)
             req = conn.getresponse()
             status = req.status
@@ -654,7 +667,7 @@ class SASsessionHTTP():
 
       return jobid
 
-   def submit(self, code: str, results: str ="html", prompt: dict = []) -> dict:
+   def submit(self, code: str, results: str ="html", prompt: dict = None, **kwargs) -> dict:
       '''
       code    - the SAS statements you want to execute 
       results - format of results, HTML is default, TEXT is the alternative
@@ -678,6 +691,9 @@ class SASsessionHTTP():
             print(results['LOG'])
             HTML(results['LST']) 
       '''
+      prompt  = prompt if prompt is not None else {}
+      printto = kwargs.pop('undo', False)
+
       #odsopen  = json.dumps("ods listing close;ods html5 (id=saspy_internal) options(bitmap_mode='inline') device=png; ods graphics on / outputfmt=png;\n")
       #odsopen  = json.dumps("ods listing close;ods html5 (id=saspy_internal) options(bitmap_mode='inline') device=svg; ods graphics on / outputfmt=png;\n")
       #odsclose = json.dumps("ods html5 (id=saspy_internal) close;ods listing;\n")
@@ -780,6 +796,20 @@ class SASsessionHTTP():
          lstd = ''
 
       self._sb._lastlog = logd
+
+      # issue 294
+      if printto:
+         conn = self.sascfg.HTTPConn; conn.connect()
+         jcode = json.dumps('proc printto;run;\n')
+         d1 = '{"code":['+jcode+']}'
+         headers={"Accept":"application/json","Content-Type":"application/vnd.sas.compute.job.request+json",
+                  "Authorization":"Bearer "+self.sascfg._token}
+         conn.request('POST', self._uri_exe, body=d1, headers=headers)
+         req = conn.getresponse()
+         status = req.status
+         resp = req.read()
+         conn.close()
+
       return dict(LOG=logd, LST=lstd)
 
    def saslog(self):
