@@ -340,8 +340,8 @@ Will use HTML5 for this SASsession.""")
       fcntl.fcntl(self.stdout, fcntl.F_SETFL, os.O_NONBLOCK)
       fcntl.fcntl(self.stderr, fcntl.F_SETFL, os.O_NONBLOCK)
 
-      rc = os.waitid(os.P_PID, self.pid, os.WEXITED | os.WNOHANG)
-      if rc != None:
+      rc = os.waitpid(self.pid, os.WNOHANG)
+      if rc[0] != 0:
          self.pid = None
          self._sb.SASpid = None
          lst = self.stdout.read1(4096)
@@ -380,13 +380,24 @@ Will use HTML5 for this SASsession.""")
             #self._asubmit(code,'text')
          sleep(1)
          if self.pid:
-            try:
-               rc = os.waitid(os.P_PID, self.pid, os.WEXITED | os.WNOHANG)
-            except (subprocess.TimeoutExpired):
+            pid = self.pid
+            x = 5
+            while True:
+               rc = os.waitpid(self.pid, os.WNOHANG)
+               if rc[0] != 0:
+                  break
+               x = x - 1
+               if x < 1:
+                  break
+               sleep(1)
+           
+            if rc[0] != 0:
+               pass
+            else:
                if self.sascfg.verbose:
                   print("SAS didn't shutdown w/in 5 seconds; killing it to be sure")
-                  ret = rc
                os.kill(self.pid, signal.SIGKILL)
+
          if self.sascfg.verbose:
             print("SAS Connection terminated. Subprocess id was "+str(self.pid))
          self.pid        = None
@@ -399,8 +410,8 @@ Will use HTML5 for this SASsession.""")
       logn   = self._logcnt(False)
       code1  = "%put E3969440A681A24088859985"+logn+";\nE3969440A681A24088859985"+logn
 
-      rc = os.waitid(os.P_PID, self.pid, os.WEXITED | os.WNOHANG)
-      if rc != None:
+      rc = os.waitpid(self.pid, os.WNOHANG)
+      if rc[0] != 0:
          self.pid = None
          self._sb.SASpid = None
          return 'SAS process has terminated unexpectedly. Pid State= '+str(rc)
@@ -421,8 +432,8 @@ Will use HTML5 for this SASsession.""")
       if self.pid == None:
          self._sb.SASpid = None
          return "No SAS process attached. SAS process has terminated unexpectedly."
-      rc = os.waitid(os.P_PID, self.pid, os.WEXITED | os.WNOHANG)
-      if rc != None:
+      rc = os.waitpid(self.pid, os.WNOHANG)
+      if rc[0] != 0:
          self.pid = None
          self._sb.SASpid = None
          return 'SAS process has terminated unexpectedly. Pid State= '+str(rc)
@@ -461,8 +472,8 @@ Will use HTML5 for this SASsession.""")
       if self.pid == None:
          self._sb.SASpid = None
          return "No SAS process attached. SAS process has terminated unexpectedly."
-      rc = os.waitid(os.P_PID, self.pid, os.WEXITED | os.WNOHANG)
-      if rc != None:
+      rc = os.waitpid(self.pid, os.WNOHANG)
+      if rc[0] != 0:
          self.pid = None
          self._sb.SASpid = None
          return 'SAS process has terminated unexpectedly. Pid State= '+str(rc)
@@ -497,8 +508,8 @@ Will use HTML5 for this SASsession.""")
       if self.pid == None:
          self._sb.SASpid = None
          return "No SAS process attached. SAS process has terminated unexpectedly."
-      rc = os.waitid(os.P_PID, self.pid, os.WEXITED | os.WNOHANG)
-      if rc != None:
+      rc = os.waitpid(self.pid, os.WNOHANG)
+      if rc[0] != 0:
          self.pid = None
          self._sb.SASpid = None
          return 'SAS process has terminated unexpectedly. Pid State= '+str(rc)
@@ -533,7 +544,7 @@ Will use HTML5 for this SASsession.""")
 
       return str(out)
 
-   def submit(self, code: str, results: str ="html", prompt: dict = None) -> dict:
+   def submit(self, code: str, results: str ="html", prompt: dict = None, **kwargs) -> dict:
       '''
       This method is used to submit any SAS code. It returns the Log and Listing as a python dictionary.
       code    - the SAS statements you want to execute
@@ -558,7 +569,8 @@ Will use HTML5 for this SASsession.""")
             print(results['LOG'])
             HTML(results['LST'])
       '''
-      prompt = prompt if prompt is not None else {}
+      prompt  = prompt if prompt is not None else {}
+      printto = kwargs.pop('undo', False)
 
       odsopen  = b"ods listing close;ods "+self.sascfg.output.encode()+ \
                  b" (id=saspy_internal) file=stdout options(bitmap_mode='inline') device=svg style="+self._sb.HTML_Style.encode()+ \
@@ -580,14 +592,15 @@ Will use HTML5 for this SASsession.""")
       pcodei   = ''
       pcodeiv  = ''
       pcodeo   = ''
+      undo     = b'proc printto;run;\n' if printto else b''
 
       if self.pid == None:
          self._sb.SASpid = None
          print("No SAS process attached. SAS process has terminated unexpectedly.")
          return dict(LOG="No SAS process attached. SAS process has terminated unexpectedly.", LST='')
 
-      rc = os.waitid(os.P_PID, self.pid, os.WEXITED | os.WNOHANG)
-      if rc != None:
+      rc = os.waitpid(self.pid, os.WNOHANG)
+      if rc[0] != 0:
          self.pid = None
          self._sb.SASpid = None
          return dict(LOG='SAS process has terminated unexpectedly. Pid State= '+str(rc), LST='')
@@ -631,14 +644,14 @@ Will use HTML5 for this SASsession.""")
       if ods:
          self.stdin.write(odsclose)
 
-      out = self.stdin.write(b'\n'+logcodei.encode(self.sascfg.encoding)+b'\n')
+      out = self.stdin.write(undo+logcodei.encode(self.sascfg.encoding)+b'\n')
       self.stdin.flush()
 
       while not done:
          try:
              while True:
-                 rc = os.waitid(os.P_PID, self.pid, os.WEXITED | os.WNOHANG)
-                 if rc is not None:
+                 rc = os.waitpid(self.pid, os.WNOHANG)
+                 if rc[0] != 0:
                      log = b''
                      try:
                         log = self.stderr.read1(4096)
@@ -668,7 +681,7 @@ Will use HTML5 for this SASsession.""")
                          if logf.count(logcodeo) >= 1:
                              bail = True
                          if not bail and bc:
-                             self.stdin.write(odsclose+logcodei.encode(self.sascfg.encoding)+b'\n')
+                             self.stdin.write(undo+odsclose+logcodei.encode(self.sascfg.encoding)+b'\n')
                              self.stdin.flush()
                              bc = False
              done = True
@@ -705,7 +718,7 @@ Will use HTML5 for this SASsession.""")
              else:
                 print('Exception ignored, continuing to process...\n')
 
-             self.stdin.write(odsclose+logcodei.encode(self.sascfg.encoding)+b'\n')
+             self.stdin.write(undo+odsclose+logcodei.encode(self.sascfg.encoding)+b'\n')
              self.stdin.flush()
 
       if ods:
@@ -763,8 +776,8 @@ Will use HTML5 for this SASsession.""")
         sleep(.25)
 
         while True:
-            rc = os.waitid(os.P_PID, self.pid, os.WEXITED | os.WNOHANG)
-            if rc is not None:
+            rc = os.waitpid(self.pid, os.WNOHANG)
+            if rc[0] != 0:
                 self.pid = None
                 self._sb.SASpid = None
                 outrc = str(rc)
@@ -800,8 +813,18 @@ Will use HTML5 for this SASsession.""")
                            bc = True
                            break
                     else:
-                        #print("******************No 'Select' or 'Press' found in lst="+lstf.decode(self.sascfg.encoding, errors='replace'))
-                        pass
+                        print("******************No 'Select' or 'Press' found. Here's what was found.")
+                        found = True
+                        print('Processing interrupt\nAttn handler Query is\n\n' + lst.decode(self.sascfg.encoding, errors='replace'))
+                        response = None
+                        while response is None:
+                           response = self.sascfg._prompt("Please enter your Response: or N/A only if there are no choices: ")
+                        self.stdin.write(response.encode(self.sascfg.encoding) + b'\n')
+                        self.stdin.flush()
+                        if response in ['N/A', '']:
+                           break
+                        found = True
+                        bc = True
             else:
                 log = self.stderr.read1(4096)
                 logf += log
@@ -1068,12 +1091,6 @@ Will use HTML5 for this SASsession.""")
                return {'Success' : False, 
                        'LOG'     : "File "+str(remotefile)+" exists and overwrite was set to False. Upload was stopped."}
 
-      try:
-         fd = open(localfile, 'rb')
-      except OSError as e:
-         return {'Success' : False, 
-                 'LOG'     : "File "+str(localfile)+" could not be opened. Error was: "+str(e)}
-
       port =  kwargs.get('port', 0)
 
       if self.sascfg.ssh and self.sascfg.rtunnel and port == 0:
@@ -1083,9 +1100,16 @@ Will use HTML5 for this SASsession.""")
       else:
          return self._upload_client(localfile, remotefile, overwrite, permission, **kwargs)
 
+      try:
+         fd = open(localfile, 'rb')
+      except OSError as e:
+         return {'Success' : False, 
+                 'LOG'     : "File "+str(localfile)+" could not be opened. Error was: "+str(e)}
+
       code = """
          filename saspydir '"""+remf+"""' recfm=F encoding=binary lrecl=1 permission='"""+permission+"""';
-         filename sock socket ':"""+str(port)+"""' server reconn=0 recfm=S encoding=binary lrecl=4096;
+         filename sock socket ':"""+str(port)+"""' server reconn=0 recfm=S lrecl=4096;
+         /* filename sock socket ':"""+str(port)+"""' server reconn=0 recfm=S encoding=binary lrecl=4096; */
 
          data _null_; nb = -1;
          infile sock nbyte=nb; 
@@ -1132,12 +1156,12 @@ Will use HTML5 for this SASsession.""")
                   sock.close()
                   fd.close()
                   break
-         except Exception as e:
+         except (KeyboardInterrupt, Exception) as e:
             sock.close()
             fd.close()
             ll = self.submit("", 'text')
             return {'Success' : False, 
-                    'LOG'     : "Download was interupted. Returning the SAS log:\n\n"+ll['LOG']}
+                    'LOG'     : "Download was interupted. Returning the SAS log:\n\n"+str(e)+"\n\n"+ll['LOG']}
         
       ll = self.submit("", 'text')
       return {'Success' : True, 
@@ -1164,12 +1188,6 @@ Will use HTML5 for this SASsession.""")
                return {'Success' : False, 
                        'LOG'     : "File "+str(remotefile)+" exists and overwrite was set to False. Upload was stopped."}
 
-      try:
-         fd = open(localfile, 'rb')
-      except OSError as e:
-         return {'Success' : False, 
-                 'LOG'     : "File "+str(localfile)+" could not be opened. Error was: "+str(e)}
-
       port =  kwargs.get('port', 0)
 
       if port==0 and self.sascfg.tunnel:
@@ -1184,6 +1202,12 @@ Will use HTML5 for this SASsession.""")
       else:
          host = ''
 
+      try:
+         fd = open(localfile, 'rb')
+      except OSError as e:
+         return {'Success' : False, 
+                 'LOG'     : "File "+str(localfile)+" could not be opened. Error was: "+str(e)}
+
       try:
          sock = socks.socket()
          if self.sascfg.tunnel:
@@ -1197,7 +1221,8 @@ Will use HTML5 for this SASsession.""")
 
       code = """
          filename saspydir '"""+remf+"""' recfm=F encoding=binary lrecl=1 permission='"""+permission+"""';
-         filename sock socket '"""+host+""":"""+str(port)+"""' recfm=S encoding=binary lrecl=4096;
+         filename sock socket '"""+host+""":"""+str(port)+"""' recfm=S lrecl=4096;
+         /* filename sock socket '"""+host+""":"""+str(port)+"""' recfm=S encoding=binary lrecl=4096; */
 
          data _null_; nb = -1;
          infile sock nbyte=nb; 
@@ -1212,6 +1237,14 @@ Will use HTML5 for this SASsession.""")
       sock.listen(1)
       self._asubmit(code, 'text')
 
+      if sel.select([sock],[],[],10)[0] == []:
+         print("error occured in SAS during upload. Check the returned LOG for issues.")
+         sock.close()
+         fd.close()
+         ll = self.submit("", 'text')
+         return {'Success' : False, 
+                 'LOG'     : "Failure in upload.\n"+ll['LOG']}
+         
       newsock = (0,0)
       try:
          newsock = sock.accept()
@@ -1234,18 +1267,18 @@ Will use HTML5 for this SASsession.""")
                sock.close()
                fd.close()
                break
-      except Exception as e:
+      except (KeyboardInterrupt, Exception) as e:
          try:
             if newsock[0]:
                newsock[0].shutdown(socks.SHUT_RDWR)
                newsock[0].close()
-         except Exception as e:
+         except:
             pass
          sock.close()
          fd.close()
          ll = self.submit("", 'text')
          return {'Success' : False, 
-                 'LOG'     : "Download was interupted. Returning the SAS log:\n\n"+ll['LOG']}
+                 'LOG'     : "Download was interupted. Returning the SAS log:\n\n"+str(e)+"\n\n"+ll['LOG']}
 
       ll = self.submit("", 'text')
       return {'Success' : True, 
@@ -1306,7 +1339,8 @@ Will use HTML5 for this SASsession.""")
 
       code = """
          filename saspydir '"""+remotefile+"""' recfm=F encoding=binary lrecl=4096;
-         filename sock socket '"""+host+""":"""+str(port)+"""' recfm=S encoding=binary;
+         filename sock socket '"""+host+""":"""+str(port)+"""' recfm=S lrecl=4096;
+         /* filename sock socket '"""+host+""":"""+str(port)+"""' recfm=S encoding=binary; */
          data _null_;
          file sock;
          infile saspydir;
@@ -1317,8 +1351,15 @@ Will use HTML5 for this SASsession.""")
       sock.listen(1)
       self._asubmit(code, 'text')
 
+      if sel.select([sock],[],[],10)[0] == []:
+         print("error occured in SAS during download. Check the returned LOG for issues.")
+         sock.close()
+         fd.close()
+         ll = self.submit("", 'text')
+         return {'Success' : False, 
+                 'LOG'     : "Failure in download.\n"+ll['LOG']}
+         
       datar = b''
-
       newsock = (0,0)
       try:
          newsock = sock.accept()
@@ -1334,15 +1375,18 @@ Will use HTML5 for this SASsession.""")
             if len(datar) > 8300:
                fd.write(datar[:8192])
                datar = datar[8192:]
-      except:
-         if newsock[0]:
-            newsock[0].shutdown(socks.SHUT_RDWR)
-            newsock[0].close()
+      except (KeyboardInterrupt, Exception) as e:
+         try:
+            if newsock[0]:
+               newsock[0].shutdown(socks.SHUT_RDWR)
+               newsock[0].close()
+         except:
+            pass
          sock.close()
          fd.close()
          ll = self.submit("filename saspydir;", 'text')
          return {'Success' : False, 
-                 'LOG'     : "Download was interupted. Returning the SAS log:\n\n"+ll['LOG']}
+                 'LOG'     : "Download was interupted. Returning the SAS log:\n\n"+str(e)+"\n\n"+ll['LOG']}
 
       newsock[0].shutdown(socks.SHUT_RDWR)
       newsock[0].close()
@@ -1613,7 +1657,8 @@ Will use HTML5 for this SASsession.""")
       if self._sb.m5dsbug:
          code = "filename sock socket '"+host+":"+str(port)+"' lrecl="+str(self.sascfg.lrecl)+" recfm=v termstr=LF;\n"
       else:
-         code = "filename sock socket '"+host+":"+str(port)+"' lrecl=1 recfm=f encoding=binary;\n"
+         code = "filename sock socket '"+host+":"+str(port)+"' recfm=S  lrecl=4096;\n"
+         #code = "filename sock socket '"+host+":"+str(port)+"' lrecl=1 recfm=f encoding=binary;\n"
 
       code += "data _null_; set "+tabname+self._sb._dsopts(dsopts)+";\n"
       for i in range(nvars):
@@ -1635,6 +1680,7 @@ Will use HTML5 for this SASsession.""")
 
       if self._sb.m5dsbug:
          rsep = colsep+rowsep+'\n'
+         csep = colsep
          code += "file sock dlm="+cdelim+";\nput "
          for i in range(nvars):
             code += " '"+varlist[i]+"'n "
@@ -1642,7 +1688,8 @@ Will use HTML5 for this SASsession.""")
                code +='\n'
          code += rdelim+";\nrun;"
       else:
-         rsep = rowsep
+         rsep = ' '+rowsep
+         csep = ' '+colsep
          code += "file sock; "
          last  = len(varlist)-1
          for i in range(nvars):
@@ -1689,7 +1736,7 @@ Will use HTML5 for this SASsession.""")
             datap = datap.decode(self.sascfg.encoding, errors='replace')
             for i in datap.split(sep=rsep):
                if i != '':
-                  r.append(tuple(i.split(sep=colsep)))
+                  r.append(tuple(i.split(sep=csep)))
 
             if len(r) > trows:
                tdf = pd.DataFrame.from_records(r, columns=varlist)
@@ -1710,14 +1757,17 @@ Will use HTML5 for this SASsession.""")
                else:
                   df = tdf
                r = []
-      except:
+      except (KeyboardInterrupt, Exception) as e:
          print("sasdata2dataframe was interupted. Trying to return the saslog instead of a data frame.")
-         if newsock[0]:
-            newsock[0].shutdown(socks.SHUT_RDWR)
-            newsock[0].close()
+         try:
+            if newsock[0]:
+               newsock[0].shutdown(socks.SHUT_RDWR)
+               newsock[0].close()
+         except Exception as e:
+            pass
          sock.close()
          ll = self.submit("", 'text')
-         return ll['LOG']
+         return str(e)+"\n\n"+ll['LOG']
 
       newsock[0].shutdown(socks.SHUT_RDWR)
       newsock[0].close()
@@ -1918,14 +1968,17 @@ Will use HTML5 for this SASsession.""")
                   break
 
                csv.write(data)
-         except:
+         except (KeyboardInterrupt, Exception) as e:
             print("sasdata2dataframe was interupted. Trying to return the saslog instead of a data frame.")
-            if newsock[0]:
-               newsock[0].shutdown(socks.SHUT_RDWR)
-               newsock[0].close()
+            try:
+               if newsock[0]:
+                  newsock[0].shutdown(socks.SHUT_RDWR)
+                  newsock[0].close()
+            except:
+               pass
             sock.close()
             ll = self.submit("", 'text')
-            return ll['LOG']
+            return str(e)+"\n\n"+ll['LOG']
 
          newsock[0].shutdown(socks.SHUT_RDWR)
          newsock[0].close()
@@ -2067,11 +2120,12 @@ Will use HTML5 for this SASsession.""")
          if self._sb.m5dsbug:
             code = "filename sock socket '"+host+":"+str(port)+"' lrecl="+str(self.sascfg.lrecl)+" recfm=v termstr=LF;\n"
          else:
-            code = "filename sock socket '"+host+":"+str(port)+"' lrecl=1 recfm=f encoding=binary;\n"
+            code = "filename sock socket '"+host+":"+str(port)+"' recfm=S lrecl=4096;\n"
+            #code = "filename sock socket '"+host+":"+str(port)+"' lrecl=1 recfm=F encoding=binary;\n"
       else:
          host = ''
          enc  = self.sascfg.encoding
-         code = "filename sock        '"+tmpcsv            +"' lrecl=1 recfm=f encoding=binary;\n"
+         code = "filename sock        '"+tmpcsv            +"' lrecl=1 recfm=F encoding=binary;\n"
 
       rdelim = "'"+'%02x' % ord(rowsep.encode(self.sascfg.encoding))+"'x"
       cdelim = "'"+'%02x' % ord(colsep.encode(self.sascfg.encoding))+"'x"
@@ -2104,6 +2158,7 @@ Will use HTML5 for this SASsession.""")
 
       if self._sb.m5dsbug:
          rsep = colsep+rowsep+'\n'
+         csep = colsep
          code += "file sock dlm="+cdelim+";\nput "
          for i in range(nvars):
             code += " '"+varlist[i]+"'n "
@@ -2111,7 +2166,8 @@ Will use HTML5 for this SASsession.""")
                code +='\n'
          code += rdelim+";\nrun;"
       else:
-         rsep = rowsep
+         rsep = ' '+rowsep
+         csep = ' '+colsep
          code += "file sock; "
          last  = len(varlist)-1
          for i in range(nvars):
@@ -2151,18 +2207,25 @@ Will use HTML5 for this SASsession.""")
                datap = data[0]+data[1]
                datar = data[2]
 
+               '''
                if not self._sb.m5dsbug:
                   csv.write(datap.decode(self.sascfg.encoding, errors='replace'))
                else:
                   csv.write(datap.decode(self.sascfg.encoding, errors='replace').replace(rsep,rowsep))
-         except:
+               '''
+               csv.write(datap.decode(self.sascfg.encoding, errors='replace').replace(rsep,rowsep))
+
+         except (KeyboardInterrupt, Exception) as e:
             print("sasdata2dataframe was interupted. Trying to return the saslog instead of a data frame.")
-            if newsock[0]:
-               newsock[0].shutdown(socks.SHUT_RDWR)
-               newsock[0].close()
+            try:
+               if newsock[0]:
+                  newsock[0].shutdown(socks.SHUT_RDWR)
+                  newsock[0].close()
+            except:
+               pass
             sock.close()
             ll = self.submit("", 'text')
-            return ll['LOG']
+            return str(e)+"\n\n"+ll['LOG']
    
          newsock[0].shutdown(socks.SHUT_RDWR)
          newsock[0].close()
